/*****************************************************************************
 * myOptimizationExample.c
 *****************************************************************************/

#include <sys/platform.h>
#include "adi_initialize.h"
#include "myOptimizationExample.h"
#include <services/gpio/adi_gpio.h>
#include <services/int/adi_int.h>
#include <stdfix.h>
#include <stdio.h>
#include <builtins.h>
#include <fract2float_conv.h>

#include "MemCopy.h"

/** 
 * If you want to use command program arguments, then place them in the following string. 
 */
char __argv_string[] = "";

void C_ITransform4x4(int *in);

int Testcase[16] = {
    0x0015,  0x0009,  0x0006,  0x0002,
    0xffe3,  0x0003,  0xfff7,  0x0000,
    0xfffc,  0xfffe,  0xffff,  0x0000,
    0xffff,  0x0000,  0x0000,  0x0000
};

/* the sum of squared difference of two vectors*/

int sumsquarediff(short* a, short* b, int size)
{
	int sum = 0;
	for(int i=0;i<size;i++){
		short tmp = a[i]-b[i];
		sum += tmp*tmp;
	}

	return sum;
}

//int sumsquarediff(short* a, short* b, int size)
//{
//	fract32 sum = 0;
//	for(int i=0;i<size;++i){
//		sum += mult_fr1x32(a[i]-b[i],a[i]-b[i]);
//	}
//
//	return sum;
//}

//int sumsquarediff(short*  a, short*  b, int size)
//{
//
//	fract * fr_a = (fract *) a;
//	fract * fr_b = (fract *) b;
//	accum sum = 0;
//
//	for(int i=0;i<size;++i){
//		sum += (fr_a[i]-fr_b[i])*(fr_a[i]-fr_b[i]);
//	}
//
//	return bitslr((long fract) sum);
//
//}


/* the sum of multiplication and shift of two vectors*/
int sumMultiShift(short* a, short* b, int size)
{
	int sum = 0;
#pragma different_banks
	for(int i=0;i<size;i++){
		int tmp = a[i]*b[i];
		sum += tmp<<1;
	}

	return sum;
}

//#pragma section("L1_data_a1")
short a[16]={1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16};
//#pragma section("L1_data_a2")
short b[16]={16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1};
//#pragma section("L1_data_a")
fract b_f[16]={0.1r,0.2r,0.3r,0.4r,0.5r,0.6r,0.7r,
		0.8r,0.9r,0.10r,0.11r,0.12r,0.13r,
		0.14r,0.15r,0.16r};
//#pragma section("L1_data_b")
fract a_f[16]={0.1r,0.2r,0.3r,0.4r,0.5r,0.6r,0.7r,
		0.8r,0.9r,0.10r,0.11r,0.12r,0.13r,
		0.14r,0.15r,0.16r};

#ifdef DMA_EXAMPLE
/***** Data buffer MDMA *****/
/*input data buffers on L3 memory */
#pragma section ("dmc_bank0")
static uint8_t SrcDataBuf0 [MY_IMG_BUF_SIZE];
static uint8_t SrcDataBuf1 [MY_IMG_BUF_SIZE];
/*output data buffer on L3 memory */
#pragma section ("dmc_bank1")
static uint8_t DestDataBuf [MY_IMG_BUF_SIZE];

/* input data buffers on L1 memory */
#pragma section ("L1_data_a1")
static uint8_t SrcDataBufA0_L1 [MY_IMG_BLK_SIZE];
#pragma section ("L1_data_a2")
static uint8_t SrcDataBufB0_L1 [MY_IMG_BLK_SIZE];
#pragma section ("L1_data_b1")
static uint8_t SrcDataBufA1_L1 [MY_IMG_BLK_SIZE];
#pragma section ("L1_data_b2")
static uint8_t SrcDataBufB1_L1 [MY_IMG_BLK_SIZE];

/* output data buffers on L1 memory */
#pragma section ("L1_data_a3")
static uint8_t DestDataBufA_L1 [MY_IMG_BLK_SIZE];
#pragma section ("L1_data_b3")
static uint8_t DestDataBufB_L1 [MY_IMG_BLK_SIZE];


extern volatile bool bMem0CopyInProgress;
extern volatile bool bMem1CopyInProgress;
extern volatile bool bMem2CopyInProgress;

static void PrepareDataBuffers (char* path1)
{
    // load two 512x512 raw images in to SrcDataBuf0 and SrcDataBuf1
	FILE *fp0 = fopen(path1,"r");

	if(!fp0){
		printf("Require image files.\n");
		fclose (fp0);
		exit(1);
	}

	if (!fread(SrcDataBuf0,MY_IMG_BUF_SIZE,1,fp0)){
		printf("Fail to load in the first file.\n");
		fclose (fp0);
		exit(1);
	}


	fclose (fp0);

	for(int i=0;i<MY_IMG_BUF_SIZE;++i){
		SrcDataBuf1[i] = SrcDataBuf0[MY_IMG_BUF_SIZE-1-i];
	}

}
ADI_DMA_RESULT MoveDataOn(int block,int index){
	ADI_DMA_RESULT eRes = ADI_DMA_SUCCESS;
	if(!index){
		eRes = StartOneShotTransmission(ADI_DMA_MEMDMA_S0, SrcDataBufA0_L1,&SrcDataBuf0[block*MY_IMG_BLK_SIZE]);
		eRes |= StartOneShotTransmission(ADI_DMA_MEMDMA_S1, SrcDataBufA1_L1,&SrcDataBuf1[block*MY_IMG_BLK_SIZE]);

	}else{
		eRes = StartOneShotTransmission(ADI_DMA_MEMDMA_S0, SrcDataBufB0_L1,&SrcDataBuf0[block*MY_IMG_BLK_SIZE]);
		eRes |= StartOneShotTransmission(ADI_DMA_MEMDMA_S1, SrcDataBufB1_L1,&SrcDataBuf1[block*MY_IMG_BLK_SIZE]);
	}

	return eRes;
}

ADI_DMA_RESULT MoveDataOff(int block,int index){
	ADI_DMA_RESULT eRes = ADI_DMA_SUCCESS;
	if(!index){
		eRes = StartOneShotTransmission(ADI_DMA_MEMDMA_S2, &DestDataBuf[block*MY_IMG_BLK_SIZE],DestDataBufA_L1);

	}else{
		eRes = StartOneShotTransmission(ADI_DMA_MEMDMA_S2, &DestDataBuf[block*MY_IMG_BLK_SIZE],DestDataBufB_L1);
	}

	return eRes;
}

void Blend(int index){

	int i;
	uint8_t *pSrc0, *pSrc1,*pDst;

	if(!index){
		pSrc0 = SrcDataBufA0_L1;
		pSrc1 = SrcDataBufA1_L1;
		pDst = DestDataBufA_L1;
		for(i=0;i<MY_IMG_BLK_SIZE;i++){
			pDst[i]=(pSrc0[i]>>1)+(pSrc0[i]>>2)+(pSrc1[i]>>2);
		}
	}else{
		pSrc0 = SrcDataBufB0_L1;
		pSrc1 = SrcDataBufB1_L1;
		pDst = DestDataBufB_L1;
		for(i=0;i<MY_IMG_BLK_SIZE;i++){
			pDst[i]=(pSrc0[i]>>1)+(pSrc0[i]>>2)+(pSrc1[i]>>2);

		}
	}

}
#endif

#pragma align 4
section ("L1_data_a") unsigned short log_table[] = {
						  #include "log_table.dat"
};

unsigned short log10i(unsigned short num){

	int hi = num>>8;
	int lo = num - (hi<<8);

	unsigned short unit,bits,val,interp;
	int offset;

	if (num == 0 || num == 1){
		return 0;
	}

	if(hi<32){
		bits = 6;
		unit = lo>>bits;
		offset = hi*4 + unit;
	}else{
		bits = 8;
		unit = lo>>bits;
		offset = hi-32 + 128;
	}

	val = (((log_table[offset+1]-log_table[offset])*(lo-unit*(1<<bits)))>>bits) + log_table[offset];

	return val;
}

int main(int argc, char *argv[])
{
	/**
	 * Initialize managed drivers and/or services that have been added to 
	 * the project.
	 * @return zero on success 
	 */

	adi_initComponents();


#ifdef DMA_EXAMPLE
	char* path="C:\\Projects\\myOptimizationExample\\lena_gray.raw";

	PrepareDataBuffers(path);

	MDMAprep();

	volatile int block = 0;
	volatile int index = 0;

	/*=============  stage 1  =============*/
	/* Load the first two blocks into the L1 memory
	 * halt the core until the transmission completes
	 */
	MoveDataOn(block,index);
	while(bMem0CopyInProgress || bMem1CopyInProgress);
	/*=============  stage 2  =============*/
	/* Load the second two blocks into the L1 memory
	 * the core is processing the first two blocks on the L1 memory
	 */
	++block;
	index ^=1;
	MoveDataOn(block,index);
	Blend(index^1);

	/*=============  stage 3  =============*/
	/* Load the third two blocks into the L1 memory
	 * the core is processing the second two blocks on the L1 memory
	 * Save the first processed block back to L3 memory
	 */

	/* The loop means:
	 * Load the n th two blocks into the L1 memory
	 * the core is processing the (n-1) th two blocks on the L1 memory
	 * Save the (n-2) th processed block back to L3 memory
	 */
	int num = MEMCOPY_NUM_DESC-2;
	while(num--){
		//wait until outgoing transmission completes
		while(bMem2CopyInProgress);
		MoveDataOff(block-1,index^1);
		//wait the incoming transmission completes
		while(bMem0CopyInProgress || bMem1CopyInProgress);
		++block;
		index ^=1;
		MoveDataOn(block,index);
		Blend(index^1);
	 }

	/*=============  stage 4  =============*/
	/*
	 * the core is processing the last two blocks on the L1 memory
	 * Save the second last processed block back to L3 memory
	 */
	while(bMem2CopyInProgress);
	MoveDataOff(block-1,index^1);
	while(bMem0CopyInProgress || bMem1CopyInProgress);
	++block;
	index ^=1;
	MoveDataOn(block,index);
	Blend(index^1);

	/*=============  stage 5  =============*/
	/*
	 * Save the last processed block back to L3 memory
	 */
	MoveDataOff(block-1,index^1);

	printf("\nDMA pipeline done \n");
#endif

	/* Begin adding your custom code here */

	int sum1 = sumMultiShift(a, b, 16);

	int sum = sumsquarediff((short *)a, (short *)b, 16);

	unsigned long start = __builtin_emuclk();


	for (int i = 0; i<5000;++i){
		C_ITransform4x4(Testcase);
	}

	unsigned long end = __builtin_emuclk();

	printf("\ntotal cycles: %d\n", end-start);


	unsigned short log_input = 6421;
	unsigned short e_val_i = log10i(log_input);
	float e_val = (float)(e_val_i>>FRACT_BITS) + fr16_to_float((e_val_i & MASK)<<(INT_BITS-1));

	printf("\n%d log value: %f\n", log_input, e_val);


}



